# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import numpy as np
from functools import reduce
from math import pi
from collections.abc import Iterable, Sequence

from .lattice import Lattice

from netket.utils import struct, deprecated_new_name
from netket.utils.types import Array, Union
from netket.utils.float import prune_zeros
from netket.utils.dispatch import dispatch

from netket.utils.group import (
    Element,
    Identity,
    PointGroup,
    Permutation,
    PermutationGroup,
)


class Translation(Permutation):
    r"""
    Custom subclass of `Permutation` that represents a lattice permutation.
    Stores translation lattice vector and generates a sensible name from it.

    The product of two `Translation`s carries the appropriate displacement vector.
    """

    def __init__(self, permutation: Array, displacement: Array):
        r"""
        Creates a `Translation` from a permutation array and a displacement vector

        Arguments:
            permutation: a 1D array listing :math:`g^{-1}(x)` for all
                :math:`0\le x < N` (i.e., `V[permutation]` permutes the
                elements of `V` as desired)
            displacement: displacement vector is units of lattice basis vectors

        Returns:
            a `Translation` object encoding the same information
        """
        super().__init__(permutation)
        self._vector = np.asarray(displacement)

    @property
    def _name(self):
        return f"Translation({self._vector.tolist()})"


@dispatch
def product(p: Translation, q: Translation):
    return Translation(p(np.asarray(q)), p._vector + q._vector)


def _ensure_iterable(x):
    """Extracts iterables given in varargs"""
    if isinstance(x[0], Iterable):
        if len(x) > 1:
            raise TypeError("Either Iterable or variable argument list expected")
        return x[0]
    else:
        return x


@struct.dataclass
class SpaceGroupBuilder(struct.Pytree):
    """
    Class to handle the space group symmetries of `Lattice`.

    Constructs `PermutationGroup`s that represent the action on a `Lattice` of
    * a geometrical point group given as a constructor argument,
    * its rotational subgroup (i.e. point group symmetries with determinant +1)
    * the translation group of the same lattice
    * and the space group that is generated as the semidirect product of
      the supplied point group and the translation group.

    Also generates space group irreps for symmetrising wave functions.
    """

    lattice: Lattice
    _point_group: PointGroup

    def __init__(self, lattice: Lattice, point_group: PointGroup):
        """
        Constructs the Space Group Builder used to concretize a point group
        which knows nothing about how many sites there are in a lattice,
        into a Permutation Group which can be used to perform calculations.

        From the point of view of group theory, you can think of this as
        taking the point group and returning a representation on the computational
        basis defined by the lattice.

        Args:
            lattice: The lattice for which to represent the point group as
                a permutation group
            point_group: The point group to be represented
        """
        self.lattice = lattice

        if not isinstance(lattice, Lattice):
            raise TypeError(
                "The lattice provided to Space Group Builder must "
                "be an instance of `Lattice`. However the lattice you provided "
                f"is of type {type(lattice)}."
            )

        if not isinstance(point_group, PointGroup):
            extra_tip = ""
            if isinstance(point_group, Element):
                extra_tip = extra_tip + (
                    "\n\n It seems you might have forgotten to wrap this "
                    "single symmetry inside of a nk.utils.group.PointGroup."
                )
            raise TypeError(
                "The point group provided to Space Group Builder must "
                "be an instance of `PointGroup`. However the lattice you provided "
                f"is of type {type(point_group)}. {extra_tip}"
            )

        point_group = point_group.replace(unit_cell=lattice.basis_vectors)
        self._point_group = point_group

    @property
    @deprecated_new_name("_point_group", reason="Consistency")
    def point_group_(self) -> PointGroup:
        """
        Deprecated: Returns the internally stored point group as a point group,
        instead of the one stored as a permutation group.
        """

        return self._point_group

    # TODO describe ordering of group elements here and later in docstring
    @struct.property_cached
    def point_group(self) -> PermutationGroup:
        """
        The point group as a `PermutationGroup` acting on the sites of `self.lattice`.
        """
        perms = []
        for p in self._point_group:
            if isinstance(p, Identity):
                perms.append(Identity())
            else:
                # note that we need the preimages in the permutation
                perm = self.lattice.id_from_position(p.preimage(self.lattice.positions))
                perms.append(Permutation(perm, name=str(p)))
        return PermutationGroup(perms, degree=self.lattice.n_nodes)

    @struct.property_cached
    def rotation_group(self) -> PermutationGroup:
        """The group of rotations (i.e. point group symmetries with determinant +1)
        as a `PermutationGroup` acting on the sites of `self.lattice`."""
        perms = []
        for p in self._point_group.rotation_group():
            if isinstance(p, Identity):
                perms.append(Identity())
            else:
                # note that we need the preimages in the permutation
                perm = self.lattice.id_from_position(p.preimage(self.lattice.positions))
                perms.append(Permutation(perm, name=str(p)))
        return PermutationGroup(perms, degree=self.lattice.n_nodes)

    def _translations_along_axis(self, axis: int) -> PermutationGroup:
        """
        The group of valid translations along an axis as a `PermutationGroup`
        acting on the sites of `self.lattice.`
        """
        if self.lattice._pbc[axis]:
            trans_list = [Identity()]
            # note that we need the preimages in the permutation
            trans_perm = self.lattice.id_from_position(
                self.lattice.positions - self.lattice.basis_vectors[axis]
            )
            vector = np.zeros(self.lattice.ndim, dtype=int)
            vector[axis] = 1
            trans_by_one = Translation(trans_perm, vector)

            for _ in range(1, self.lattice.extent[axis]):
                trans_list.append(trans_list[-1] @ trans_by_one)

            return PermutationGroup(trans_list, degree=self.lattice.n_nodes)
        else:
            return PermutationGroup([Identity()], degree=self.lattice.n_nodes)

    @struct.property_cached
    def _full_translation_group(self) -> PermutationGroup:
        """
        The group of valid translations of `self.lattice` as a `PermutationGroup`
        acting on the sites of the same.
        """
        return reduce(
            PermutationGroup.__matmul__,
            [self._translations_along_axis(i) for i in range(self.lattice.ndim)],
        )

    def translation_group(
        self, axes: Union[int, Sequence[int]] | None = None
    ) -> PermutationGroup:
        """
        The group of valid translations of `self.lattice` as a `PermutationGroup`
        acting on the sites of the same.
        """
        if axes is None:
            return self._full_translation_group
        elif isinstance(axes, int):
            return self._translations_along_axis(axes)
        else:
            return reduce(
                PermutationGroup.__matmul__,
                [self._translations_along_axis(i) for i in axes],
            )

    @struct.property_cached
    def space_group(self) -> PermutationGroup:
        """
        The space group generated by `self.point_group` and `self.translation_group`.
        """
        return self._full_translation_group @ self.point_group

    def _little_group_index(self, k: Array) -> Array:
        """
        Returns the indices of the elements of the little group corresponding to
        wave vector `k`.
        """
        # calculate k' = p(k) for all p in the point group
        big_star = np.tensordot(self._point_group.matrices(), k, axes=1)
        big_star = self.lattice.to_reciprocal_lattice(big_star) % self.lattice.extent
        # should test for pbc before taking the modulus, but the only valid wave
        # vector for non-pbc axes is 0 and 0 % anything == 0

        # assumes point_group_[0] is the identity
        is_in_little_group = np.all(big_star == big_star[0], axis=1)
        return np.arange(len(self._point_group))[is_in_little_group]

    def little_group(self, *k: Array) -> PointGroup:
        """
        Returns the little co-group corresponding to wave vector *k*.
        This is the subgroup of the point group that leaves *k* invariant.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            the little co-group as a `PointGroup`
        """
        k = _ensure_iterable(k)
        return PointGroup(
            [self._point_group[i] for i in self._little_group_index(k)],
            ndim=self._point_group.ndim,
            unit_cell=self.lattice.basis_vectors,
        )

    def _little_group_irreps(self, k: Array, divide: bool = False) -> Array:
        """
        Returns the character table of the little group embedded in the full point
        group. Symmetries outside the little group get 0.
        If `divide` is `True`, the result gets divided by the size of the little group.
        This is convenient when calculating space group irreps.
        """
        idx = self._little_group_index(k)
        CT = self.little_group(k).character_table()
        CT_full = np.zeros((CT.shape[0], len(self._point_group)), dtype=CT.dtype)
        CT_full[:, idx] = CT
        return CT_full / idx.size if divide else CT_full

    def space_group_irreps(self, *k: Array) -> Array:
        """
        Returns the portion of the character table of the full space group corresponding
        to the star of the wave vector *k*.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            An array `CT` listing the characters for a number of irreps of the
            space group.
            `CT[i]` for each `i` gives a distinct irrep, each corresponding to
            `self.little_group(k).character_table[i].
            `CT[i,j]` gives the character of `self.space_group[j]` in the same.
        """
        k = _ensure_iterable(k)
        # Wave vectors
        big_star_Cart = np.tensordot(self._point_group.matrices(), k, axes=1)
        big_star = self.lattice.to_reciprocal_lattice(big_star_Cart) * (
            2 * pi / self.lattice.extent
        )
        # Little-group-irrep factors
        # Conjugacy_table[g,p] lists p^{-1}gp, so point_group_factors[i,:,p]
        #     of irrep #i for the little group of p(k) is the equivalent
        # Phase factor for non-symmorphic symmetries is exp(-i w_g . p(k))
        point_group_factors = self._little_group_irreps(k, divide=True)[
            :, self._point_group.conjugacy_table
        ] * np.exp(
            -1j
            * np.tensordot(
                self._point_group.translations(), big_star_Cart, axes=(-1, -1)
            )
        )
        # Translational factors
        trans_factors = []
        for axis in range(self.lattice.ndim):
            n_trans = self.lattice.extent[axis] if self.lattice.pbc[axis] else 1
            factors = np.exp(-1j * np.outer(np.arange(n_trans), big_star[:, axis]))
            shape = (
                [1] * axis
                + [n_trans]
                + [1] * (self.lattice.ndim - 1 - axis)
                + [len(self._point_group)]
            )
            trans_factors.append(factors.reshape(shape))
        trans_factors = reduce(np.multiply, trans_factors).reshape(
            -1, len(self._point_group)
        )

        # Multiply the factors together and sum over the "p" PGSymmetry axis
        # Translations are more major than point group operations
        result = np.einsum(
            "igp, tp -> itg", point_group_factors, trans_factors
        ).reshape(point_group_factors.shape[0], -1)
        return prune_zeros(result)

    def one_arm_irreps(self, *k: Array) -> Array:
        """
        Returns the portion of the character table of the full space group corresponding
        to the star of the wave vector *k*, projected onto *k* itself.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            An array `CT` listing the projected characters for a number of irreps of
            the space group.
            `CT[i]` for each `i` gives a distinct irrep, each corresponding to
            `self.little_group(k).character_table[i].
            `CT[i,j]` gives the character of `self.space_group[j]` in the same.
        """
        # Convert k to reciprocal lattice vectors
        k = _ensure_iterable(k)
        # Little-group irrep factors
        # Phase factor for non-symmorphic symmetries is exp(-i w_g . p(k))
        point_group_factors = self._little_group_irreps(k) * np.exp(
            -1j * (self._point_group.translations() @ k)
        )
        # Translational factors
        trans_factors = []
        for axis in range(self.lattice.ndim):
            n_trans = self.lattice.extent[axis] if self.lattice.pbc[axis] else 1
            factors = np.exp(-1j * k[axis] * np.arange(n_trans))
            shape = [1] * axis + [n_trans] + [1] * (self.lattice.ndim - 1 - axis)
            trans_factors.append(factors.reshape(shape))
        trans_factors = reduce(np.multiply, trans_factors).ravel()

        # Multiply the factors together
        # Translations are more major than point group operations
        result = np.einsum("ig, t -> itg", point_group_factors, trans_factors).reshape(
            point_group_factors.shape[0], -1
        )
        return prune_zeros(result)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import numpy as jnp

from netket.hilbert.custom_hilbert import CustomHilbert
from netket.utils.dispatch import dispatch


@dispatch
def random_state(hilb: CustomHilbert, key, batches: int, *, dtype):
    if not hilb.is_finite or hilb.constrained:
        raise NotImplementedError()

    # Default version for discrete hilbert spaces without constraints.
    # More specialized initializations can be defined in the derived classes.

    σ = jax.random.choice(
        key,
        jnp.asarray(hilb.local_states, dtype=dtype),
        shape=(batches, hilb.size),
        replace=True,
    )
    return jnp.asarray(σ, dtype=dtype)


@dispatch
def flip_state_scalar(hilb: CustomHilbert, key, σ, indx):
    local_states = jnp.asarray(hilb.local_states)

    rs = jax.random.randint(key, shape=(), minval=0, maxval=len(hilb.local_states) - 1)

    new_val = local_states[rs + (local_states[rs] >= σ[indx])]
    return σ.at[indx].set(new_val), σ[indx]


@dispatch
def flip_state_batch(hilb: CustomHilbert, key, σ, indxs):
    n_batches = σ.shape[0]

    local_states = jnp.asarray(hilb.local_states)

    rs = jax.random.randint(
        key, shape=(n_batches,), minval=0, maxval=len(hilb.local_states) - 1
    )

    def scalar_update_fun(σ, indx, rs):
        new_val = local_states[rs + (local_states[rs] >= σ[indx])]
        return σ.at[indx].set(new_val), σ[indx]

    return jax.vmap(scalar_update_fun, in_axes=(0, 0, 0), out_axes=0)(σ, indxs, rs)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "driver",
    "dynamics",
    "sampler",
    "vqs",
    "TDVP",
    "models",
    "hilbert",
    "operator",
    "logging",
    "observable",
]

from . import hilbert
from . import operator
from . import driver
from . import dynamics
from . import sampler
from . import models
from . import vqs
from . import logging
from . import qsr
from . import observable

from .driver import TDVP
from .qsr import QSR

from netket.utils import _hide_submodules

_hide_submodules(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._abstract_operator import AbstractOperator

from ._discrete_operator import DiscreteOperator
from ._discrete_operator_jax import DiscreteJaxOperator

from ._pauli_strings import PauliStrings, PauliStringsJax
from ._local_operator import LocalOperator, LocalOperatorJax
from ._graph_operator import GraphOperator
from ._lazy import Adjoint, Transpose, Squared
from ._heisenberg import Heisenberg
from ._ising import Ising, IsingJax
from ._bose_hubbard import BoseHubbard, BoseHubbardJax

from ._abstract_super_operator import AbstractSuperOperator
from ._local_liouvillian import LocalLiouvillian

from ._continuous_operator import ContinuousOperator
from ._kinetic import KineticEnergy
from ._potential import PotentialEnergy
from ._sumoperators import SumOperator

from . import spin, boson

from netket.utils import _auto_export

_auto_export(__name__)



from collections.abc import Callable

import jax
from jax import numpy as jnp
from flax import linen as nn
from netket.utils.types import NNInitFunc, DType
from jax.nn.initializers import (
    zeros,
    lecun_normal,
)

from .mlp import MLP


def _process_features(features) -> tuple[tuple[int, ...] | None, int | None]:
    """
    Convert some inputs to a consistent format of features.
    Returns hidden dimensions and output dimensions of the MLP separately.
    """
    if features is None:
        feat, out = None, None
    elif isinstance(features, int):
        feat, out = None, features
    elif len(features) == 0:
        feat, out = None, None
    elif len(features) == 1:
        feat, out = None, features[0]
    else:
        feat, out = tuple(features[:-1]), features[-1]
    return feat, out


class DeepSetMLP(nn.Module):
    r"""Implements the DeepSets architecture, which is permutation invariant
    and is suitable for the encoding of bosonic systems.

    .. math::

        f(x_1,...,x_N) = \rho\left(\sum_i \phi(x_i)\right)

    The input shape must have an axis that is reshaped to `(..., N, D)`, where we pool over N.

    """

    features_phi: int | tuple[int, ...] | None = None
    """
    Number of features in each layer for phi network.
    When features_phi is None, no phi network is created.
    """
    features_rho: int | tuple[int, ...] | None = None
    """
    Number of features in each layer for rho network.
    Should include final dimension of the network.
    When features_rho is None, no rho network is created.
    """

    param_dtype: DType = jnp.float64
    """The dtype of the weights."""

    hidden_activation: Callable | None = jax.nn.gelu
    """The nonlinear activation function between hidden layers."""
    output_activation: Callable | None = None
    """The nonlinear activation function at the output layer."""

    pooling: Callable = jnp.sum
    """The pooling operation to be used after the phi-transformation"""

    use_bias: bool = True
    """if True uses a bias in all layers."""

    kernel_init: NNInitFunc = lecun_normal()
    """Initializer for the Dense layer matrix"""
    bias_init: NNInitFunc = zeros
    """Initializer for the hidden bias"""
    precision: jax.lax.Precision | None = None
    """numerical precision of the computation see :class:`jax.lax.Precision` for details."""

    def setup(self):
        def _create_mlp(features, output_activation, name):
            hidden_dims, out_dim = _process_features(features)
            if out_dim is None:
                return None
            else:
                return MLP(
                    output_dim=out_dim,
                    hidden_dims=hidden_dims,
                    param_dtype=self.param_dtype,
                    hidden_activations=self.hidden_activation,
                    output_activation=output_activation,
                    use_hidden_bias=self.use_bias,
                    use_output_bias=self.use_bias,
                    kernel_init=self.kernel_init,
                    name=name,
                )

        self.phi = _create_mlp(self.features_phi, self.hidden_activation, "ds_phi")
        self.rho = _create_mlp(self.features_rho, self.output_activation, "ds_rho")

        if self.pooling is None:
            raise ValueError("Must specifyc pooling function for a DeepSet")

    @nn.compact
    def __call__(self, x):
        """The input shape must have an axis that is reshaped to (..., N, D), where we pool over N."""
        if x.ndim < 2:
            raise ValueError(
                f"input of deepset should have shape (..., N, D), but got {x.shape}"
            )

        if self.phi:
            x = self.phi(x)
        x = self.pooling(x, axis=-2)
        if self.rho:
            x = self.rho(x)

        return x



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np

import jax
import jax.numpy as jnp

from equinox import error_if

from netket.errors import InvalidConstraintInterface, UnhashableConstraintError
from netket.jax import sharding
from netket.utils import StaticRange, warn_deprecation
from netket.utils.types import Array

from .discrete_hilbert import DiscreteHilbert
from .constraint import DiscreteHilbertConstraint
from .index import (
    HilbertIndex,
    UniformTensorProductHilbertIndex,
    optimalConstrainedHilbertindex,
)


def check_and_deprecate_constraint(
    constraint,
    constraint_fn,
):
    __tracebackhide__ = True

    if constraint is not None and constraint_fn is not None:
        raise ValueError(
            "Cannot specify at the same time `constraint` and `constraint_fn`, which"
            "has been deprecated. Please remove the latter."
        )
    elif constraint_fn is not None and constraint is None:
        constraint = constraint_fn
        warn_deprecation(
            "The keyword argument `constraint_fn` was renamed to `constraint` and "
            "deprecated. Moreover, you can no longer pass a function, but must specify a class "
            "according to the documentation."
        )

    if constraint is None:
        return constraint

    # now
    if not isinstance(constraint, DiscreteHilbertConstraint):
        raise InvalidConstraintInterface()

    # Check that the constraint is well behaved and is jax-friendly hashable
    try:
        leaves, struct = jax.tree_util.tree_flatten(constraint)
        if not len(leaves) == 0:
            raise TypeError
        hash(constraint)
        constraint == constraint

    except Exception as err:
        raise UnhashableConstraintError(constraint) from err

    return constraint


class HomogeneousHilbert(DiscreteHilbert):
    r"""The Abstract base class for homogeneous hilbert spaces.

    This class should only be subclassed and should not be instantiated directly.

    .. note::

        To override the logic to index into constrained hilbert spaces, it is
        possible to use an informal interface built on top of non-public
        indexing classes.

        In particular, you can override the following properties and methods:

        - Do not specify the :code:`constraint_fn` keyword argument when
          calling the init method of this abstract class.
        - Override the property :py:attr:`~nk.hilbert.HomogeneousHilbert.constrained`,
          to return `True` or `False` depending on your own logic.
        - Override the property :py:attr:`~nk.hilbert.HomogeneousHilbert._hilbert_index`
          to return an hilbert index object (see the discussion in the source code of
          the folder :code:`netket/hilbert/index/__init__.py`).

    """

    def __init__(
        self,
        local_states: StaticRange,
        N: int = 1,
        *,
        constraint: DiscreteHilbertConstraint | None = None,
        constraint_fn: Callable | None = None,
    ):
        r"""
        Constructs a new :class:`~netket.hilbert.HomogeneousHilbert` given a list of
        eigenvalues of the states and a number of sites, or modes, within this
        hilbert space.

        This method should only be called from the subclasses `__init__` method.

        Args:
            local_states: :class:`~netket.utils.StaticRange` object describing the
                numbers used to encode the local degree of freedom of this Hilbert
                Space.
            N: Number of modes in this hilbert space (default 1).
            constraint: A callable class specifying constraints on the quantum numbers.
                Given a batch of quantum numbers it should return a vector of bools
                specifying whether those states are valid or not. It must be an
                hashable subclass of :class:`netket.hilbert.constraint.DiscreteHilbertConstraint`.

        """
        assert isinstance(N, int)

        if not (isinstance(local_states, StaticRange) or local_states is None):
            raise TypeError("local_states must be a StaticRange.")

        self._local_states = local_states
        self._local_size = len(local_states)
        self._is_finite = self._local_size < np.iinfo(np.intp).max

        self._constraint = check_and_deprecate_constraint(constraint, constraint_fn)

        self._hilbert_index_ = None

        shape = tuple(self._local_size for _ in range(N))
        super().__init__(shape=shape)

    @property
    def size(self) -> int:
        r"""The total number number of degrees of freedom."""
        return len(self.shape)

    @property
    def local_size(self) -> int:
        r"""Size of the local degrees of freedom that make the total hilbert space."""
        return self._local_size

    def size_at_index(self, i: int) -> int:
        return self.local_size

    @property
    def local_states(self) -> list[float] | None:
        r"""A list of discrete local quantum numbers.
        If the local states are infinitely many, None is returned."""
        if self.is_finite:
            return list(self._local_states)
        return self._local_states

    def states_at_index(self, i: int):
        return self.local_states

    @property
    def constraint(self) -> DiscreteHilbertConstraint:
        """
        The constraint inflicted upon this poor Hilbert space.

        Instead of all possible combinations of local states, only those satisfying
        the constraint are part of this Hilbert space.
        """
        return self._constraint

    @property
    def n_states(self) -> int:
        r"""The total dimension of the many-body Hilbert space.
        Throws an exception iff the space is not indexable."""
        if not self.is_indexable:
            raise RuntimeError("The hilbert space is too large to be indexed.")
        return self._hilbert_index.n_states

    def states_to_local_indices(self, x: Array):
        r"""Returns a tensor with the same shape of `x`, where all local
        values are converted to indices in the range `0...self.shape[i]`.
        This function is guaranteed to be jax-jittable.

        For the `Fock` space this returns `x`, but for other hilbert spaces
        such as `Spin` this returns an array of indices.

        .. warning::
            This function is experimental. Use at your own risk.

        Args:
            x: a tensor containing samples from this hilbert space

        Returns:
            a tensor containing integer indices into the local hilbert
        """
        return self._local_states.states_to_numbers(x, dtype=np.int32)

    def local_indices_to_states(self, x: Array, dtype=None):
        r"""
        Converts a tensor of integers to the corresponding local_values in
        this hilbert space.

        Equivalent to

        .. code::py

            hilbert.local_states[x]

        The input last dimension must match the size of this Hilbert space.
        This function can be jax-jitted.

        Args:
            x: a tensor with integer dtype and whose last dimension matches
                the size of this Hilbert space.

        Returns:
            a tensor with the same shape as the input, and values corresponding
            to the local_state indexed by the input tensor `x`.

        """
        return self._local_states.numbers_to_states(x, dtype=dtype)

    @property
    def is_finite(self) -> bool:
        r"""Whether the local hilbert space is finite."""
        # De-facto, all Homogeneous Hilbert spaces, as they have internally
        # a StaticRange, are factually finite. So, even when we say this is
        # False we are lying because it's technically always Finite...
        #
        # Do we still need this flag?
        return self._is_finite

    @property
    def constrained(self) -> bool:
        r"""The hilbert space does not contain `prod(hilbert.shape)`
        basis states.

        Typical constraints are population constraints (such as fixed
        number of bosons, fixed magnetization...) which ensure that
        only a subset of the total unconstrained space is populated.

        Typically, objects defined in the constrained space cannot be
        converted to QuTiP or other formats.
        """
        return self.constraint is not None

    def _numbers_to_states(self, numbers: np.ndarray) -> np.ndarray:
        return self._hilbert_index.numbers_to_states(numbers)

    def _states_to_numbers(self, states: np.ndarray):
        states = jnp.asarray(states)

        if self.is_finite:
            start = self._local_states.start
            end = start + self._local_states.step * self._local_states.length
            if self._local_states.step < 0:
                start, end = end, start
                start = start - self._local_states.step
                end = end - self._local_states.step

            # equinox.error_if is broken under shard_map.
            # If we are using shard map, we skip this check
            if sharding.SHARD_MAP_STACK_LEVEL == 0:
                states = error_if(
                    states,
                    (states < start).any() | (states >= end).any(),
                    "States outside the range of allowed states.",
                )

        if self.constrained:
            if sharding.SHARD_MAP_STACK_LEVEL == 0:
                states = error_if(
                    states,
                    ~self.constraint(states).all(),
                    "States do not fulfill constraint.",
                )

        return self._hilbert_index.states_to_numbers(states)

    def all_states(self) -> np.ndarray:
        r"""Returns all valid states of the Hilbert space.

        Throws an exception if the space is not indexable.

        Returns:
            A (n_states x size) batch of states. this corresponds
            to the pre-allocated array if it was passed.
        """
        if not self.is_indexable:  # includes call to _setup
            raise RuntimeError("The hilbert space is too large to be indexed.")

        return self._hilbert_index.all_states()

    @property
    def _hilbert_index(self) -> HilbertIndex:
        """
        The `self._hilbert_index` is a lazily constructed object used to index into homogeneous Hilbert spaces.

        This indexing object implements the logic for `number_to_states`, `states_to_numbers` and `n_states`,
        as well as the handling of constraints if necessary.
        """
        if self._hilbert_index_ is None:
            if not self.constrained:
                index = UniformTensorProductHilbertIndex(self._local_states, self.size)
            else:
                index = optimalConstrainedHilbertindex(
                    self._local_states, self.size, self.constraint
                )
            self._hilbert_index_ = index
        return self._hilbert_index_

    @property
    def is_indexable(self) -> bool:
        """Whether the space can be indexed with an integer"""
        if not self.is_finite:
            return False
        return self._hilbert_index.is_indexable

    def __repr__(self):
        constr = f", constrained={self.constrained}" if self.constrained else ""

        clsname = type(self).__name__
        return f"{clsname}(local_size={self._local_size}, N={self.size}{constr})"

    @property
    def _attrs(self):
        return (
            self.size,
            self.local_size,
            self._local_states,
            self.constrained,
            self._constraint,
        )



# Copyright 2023  The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

from functools import partial
from textwrap import dedent
import warnings

import jax
import jax.numpy as jnp
import jax.scipy as jsp

from netket import jax as nkjax
from netket import stats as nkstats
from netket.driver import AbstractVariationalDriver
from netket.errors import UnoptimalSRtWarning
from netket.jax import sharding
from netket.operator import AbstractOperator
from netket.utils import mpi, timing
from netket.utils.types import ScalarOrSchedule, Optimizer, PyTree
from netket.vqs import MCState

from jax.flatten_util import ravel_pytree


@timing.timed
@partial(jax.jit, static_argnames=("mode", "solver_fn"))
def SRt(
    O_L, local_energies, diag_shift, *, mode, solver_fn, e_mean=None, params_structure
):
    """
    For more details, see `https://arxiv.org/abs/2310.05715'. In particular,
    the following parallel implementation is described in Appendix "Distributed SR computation".
    """
    N_params = O_L.shape[-1]
    N_mc = O_L.shape[0] * mpi.n_nodes

    local_energies = local_energies.flatten()

    if e_mean is None:
        e_mean = nkstats.mean(local_energies)
    de = jnp.conj(local_energies - e_mean).squeeze()

    # * in this case O_L should be padded with zeros
    assert (N_params % mpi.n_nodes) == 0

    O_L = O_L / N_mc**0.5
    dv = -2.0 * de / N_mc**0.5

    if mode == "complex":
        # Concatenate the real and imaginary derivatives of the ansatz
        # O_L = jnp.concatenate((O_L[:, 0], O_L[:, 1]), axis=0)
        O_L = jnp.transpose(O_L, (1, 0, 2)).reshape(-1, O_L.shape[-1])

        dv = jnp.concatenate((jnp.real(dv), -jnp.imag(dv)), axis=-1)
    elif mode == "real":
        dv = dv.real
    else:
        raise NotImplementedError()

    # twons, (np, n_nodes) -> twons, np, n_nodes
    O_LT = O_L.reshape(O_L.shape[0], -1, mpi.n_nodes)
    # twons, np, n_nodes -> n_nodes, twons, np
    O_LT = jnp.moveaxis(O_LT, -1, 0)

    dv, token = mpi.mpi_gather_jax(dv)
    dv = dv.reshape(-1, *dv.shape[2:])
    O_LT, token = mpi.mpi_alltoall_jax(O_LT, token=token)

    # proc, twons, np -> (proc, twons) np
    O_LT = O_LT.reshape(-1, O_LT.shape[-1])

    matrix, token = mpi.mpi_reduce_sum_jax(O_LT @ O_LT.T, token=token)
    matrix_side = matrix.shape[-1]  # * it can be Ns or 2*Ns, depending on mode

    if mpi.rank == 0:
        matrix = matrix + diag_shift * jnp.eye(
            matrix_side
        )  # * shift diagonal regularization
        aus_vector = solver_fn(matrix, dv)
        # some solvers return a tuple, some others do not.
        # We check and try to support both
        if isinstance(aus_vector, tuple):
            aus_vector, _ = aus_vector

        aus_vector = aus_vector.reshape(mpi.n_nodes, -1)
        aus_vector, token = mpi.mpi_scatter_jax(aus_vector, token=token)
    else:
        shape = jnp.zeros((int(matrix_side / mpi.n_nodes),), dtype=jnp.float64)
        aus_vector, token = mpi.mpi_scatter_jax(shape, token=token)

    updates = O_L.T @ aus_vector
    updates, token = mpi.mpi_allreduce_sum_jax(updates, token=token)

    # If complex mode and we have complex parameters, we need
    # To repack the real coefficients in order to get complex updates
    if mode == "complex" and nkjax.tree_leaf_iscomplex(params_structure):
        np = updates.shape[-1] // 2
        updates = updates[:np] + 1j * updates[np:]

    return -updates


inv_default_solver = lambda A, b: jnp.linalg.inv(A) @ b
linear_solver = lambda A, b: jsp.linalg.solve(A, b, assume_a="pos")


@jax.jit
def _flatten_samples(x):
    return x.reshape(-1, x.shape[-1])


class VMC_SRt(AbstractVariationalDriver):
    r"""
    Energy minimization using Variational Monte Carlo (VMC) and the kernel
    formulation of Stochastic Reconfiguration (SR). This approach lead to
    *exactly* the same parameter updates of the standard SR with a
    diagonal shift regularization. For this reason, it is equivalent to the standard
    nk.driver.VMC with the preconditioner nk.optimizer.SR(solver=netket.optimizer.solver.solvers.solve)).
    In the kernel SR framework, the updates of the parameters can be written as:

    .. math::
        \delta \theta = \tau X(X^TX + \lambda \mathbb{I}_{2M})^{-1} f,

    where :math:`X \in R^{P \times 2M}` is the concatenation of the real and imaginary part
    of the centered Jacobian, with P the number of parameters and M the number of samples.
    The vector f is the concatenation of the real and imaginary part of the centered local
    energy. Note that, to compute the updates, it is sufficient to invert an :math:`M\times M` matrix
    instead of a :math:`P\times P` one. As a consequence, this formulation is useful
    in the typical deep learning regime where :math:`P \gg M`.

    See `R.Rende, L.L.Viteritti, L.Bardone, F.Becca and S.Goldt <https://arxiv.org/abs/2310.05715>`_
    for a detailed description of the derivation. A similar result can be obtained by minimizing the
    Fubini-Study distance with a specific constrain, see `A.Chen and M.Heyl <https://arxiv.org/abs/2302.01941>`_
    for details.
    """

    def __init__(
        self,
        hamiltonian: AbstractOperator,
        optimizer: Optimizer,
        *,
        diag_shift: ScalarOrSchedule,
        linear_solver_fn: Callable[[jax.Array, jax.Array], jax.Array] = linear_solver,
        jacobian_mode: str | None = None,
        variational_state: MCState = None,
    ):
        """
        Initializes the driver class.

        Args:
            hamiltonian: The Hamiltonian of the system.
            optimizer: Determines how optimization steps are performed given the
                    bare energy gradient.
            diag_shift: The diagonal shift of the stochastic reconfiguration matrix.
                        Typical values are 1e-4 ÷ 1e-3. Can also be an optax schedule.
            hamiltonian: The Hamiltonian of the system.
            linear_solver_fn: Callable to solve the linear problem associated to the
                              updates of the parameters
            jacobian_mode: The mode used to compute the jacobian of the variational state. Can be `'real'`
                    or `'complex'` (defaults to the dtype of the output of the model).
            variational_state: The :class:`netket.vqs.MCState` to be optimised. Other
                variational states are not supported.
        """
        super().__init__(variational_state, optimizer, minimized_quantity_name="Energy")

        if variational_state.hilbert != hamiltonian.hilbert:
            raise TypeError(
                dedent(
                    f"""the variational_state has hilbert space {variational_state.hilbert}
                    (this is normally defined by the hilbert space in the sampler), but
                    the hamiltonian has hilbert space {hamiltonian.hilbert}.
                    The two should match.
                    """
                )
            )

        if self.state.n_parameters % sharding.device_count() != 0:
            raise NotImplementedError(
                f"""
                VMC_SRt requires a network with a number of parameters
                multiple of the number of MPI devices/ranks in use.

                You have a network with {self.state.n_parameters}, but
                there are {sharding.device_count()} MPI ranks in use.

                To fix this, either add some 'fake' parameters to your
                network, or change the number of MPI nodes, or contribute
                some padding logic to NetKet!
                """
            )

        if self.state.n_parameters < self.state.n_samples:
            warnings.warn(
                UnoptimalSRtWarning(self.state.n_parameters, self.state.n_samples),
                UserWarning,
                stacklevel=2,
            )

        self._ham = hamiltonian.collect()  # type: AbstractOperator
        self._dp: PyTree = None

        self.diag_shift = diag_shift
        self.jacobian_mode = jacobian_mode
        self._linear_solver_fn = linear_solver_fn

        self._params_structure = jax.tree_util.tree_map(
            lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), self.state.parameters
        )
        if not nkjax.tree_ishomogeneous(self._params_structure):
            raise ValueError(
                "SRt only supports neural networks with all real or all complex "
                "parameters. Hybrid structures are not yet supported (but we would welcome "
                "contributions. Get in touch with us!)"
            )

        _, unravel_params_fn = ravel_pytree(self.state.parameters)
        self._unravel_params_fn = jax.jit(unravel_params_fn)

    @property
    def jacobian_mode(self) -> str:
        """
        The mode used to compute the jacobian of the variational state. Can be `'real'`
        or `'complex'`.

        Real mode truncates imaginary part of the wavefunction, while `complex` does not.
        This internally uses :func:`netket.jax.jacobian`. See that function for a more
        complete documentation.
        """
        return self._jacobian_mode

    @jacobian_mode.setter
    def jacobian_mode(self, mode: str | None):
        if mode is None:
            mode = nkjax.jacobian_default_mode(
                self.state._apply_fun,
                self.state.parameters,
                self.state.model_state,
                self.state.samples,
                warn=False,
            )

        if mode not in ["complex", "real"]:
            raise ValueError(
                "`jacobian_mode` only supports 'real' for real-valued wavefunctions and"
                "'complex'.\n\n"
                "`holomorphic` is not yet supported, but could be contributed in the future."
            )
        self._jacobian_mode = mode

    @timing.timed
    def _forward_and_backward(self):
        """
        Performs a number of VMC optimization steps.

        Args:
            n_steps (int): Number of steps to perform.
        """

        self.state.reset()

        # Compute the local energy estimator and average Energy
        local_energies = self.state.local_estimators(self._ham)

        self._loss_stats = nkstats.statistics(local_energies)

        samples = _flatten_samples(self.state.samples)
        jacobians = nkjax.jacobian(
            self.state._apply_fun,
            self.state.parameters,
            samples,
            self.state.model_state,
            mode=self.jacobian_mode,
            dense=True,
            center=True,
            chunk_size=self.state.chunk_size,
        )  # jacobians is centered

        diag_shift = self.diag_shift
        if callable(self.diag_shift):
            diag_shift = diag_shift(self.step_count)

        updates = SRt(
            jacobians,
            local_energies,
            diag_shift,
            mode=self.jacobian_mode,
            solver_fn=self._linear_solver_fn,
            e_mean=self._loss_stats.Mean,
            params_structure=self._params_structure,
        )

        self._dp = self._unravel_params_fn(updates)

        return self._dp

    @property
    def energy(self) -> nkstats.Stats:
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.
        """
        return self._loss_stats

    def __repr__(self):
        return (
            "Vmc_SRt("
            + f"\n  step_count = {self.step_count},"
            + f"\n  state = {self.state})"
        )



import itertools
import math
import numpy as np

import jax
import jax.numpy as jnp

from netket.utils import struct, StaticRange
from netket.utils.types import Array

from netket.hilbert.constraint import SumConstraint

from .base import HilbertIndex, is_indexable
from .uniform_tensor import UniformTensorProductHilbertIndex
from .unconstrained import LookupTableHilbertIndex
from .constrained_generic import ConstrainedHilbertIndex, optimalConstrainedHilbertindex


@optimalConstrainedHilbertindex.dispatch
def optimalConstrainedHilbertindex(local_states, size, constraint: SumConstraint):
    # If we have a constraint, we tentatively construct a specialised Hilbert index for that particular constraint.
    # If this specialised indexer object exists, we check whether it is more efficient than the generic
    # ConstrainedHilbertIndex one. If it is more efficient, we use it, otherwise we keep the generic one.
    generic_index = ConstrainedHilbertIndex(
        UniformTensorProductHilbertIndex(local_states, size), constraint
    )
    specialized_index = SumConstrainedHilbertIndex(
        local_states, size, constraint.sum_value
    )
    if specialized_index.n_states_bound < generic_index.n_states_bound:
        return specialized_index
    else:
        return generic_index


@struct.dataclass
class SumConstrainedHilbertIndex(HilbertIndex):
    """
    Specialized implementation for a constrained space with a SumConstraint.
    Does not require the unconstrained space to be indexable.
    """

    range: StaticRange = struct.field(pytree_node=True)
    size: int = struct.field(pytree_node=False)
    sum_value: int = struct.field(pytree_node=False)

    @property
    def shape(self):
        return (self.range.length,) * self.size

    @property
    def n_particles(self):
        # sum_value: e.g. total_sz
        # convert the constraint to a constraint for the number of particles
        # of a fock space with staes in the range [0,1,...,n]
        return round((self.sum_value - self.range.start * self.size) / self.range.step)

    @property
    def n_states(self):
        if self._n_max == 1:
            return math.comb(self.size, self.n_particles)
        else:
            return self._lookup_table.n_states

    @property
    def _n_max(self):
        return max(self.shape) - 1

    def states_to_numbers(self, states: Array) -> Array:
        return self._lookup_table.states_to_numbers(states)

    def numbers_to_states(self, numbers: Array):
        return self._lookup_table.numbers_to_states(numbers)

    def all_states(self):
        return self._lookup_table.all_states()

    @jax.jit
    def _compute_all_states(self):
        if self.n_particles == 0:
            return jnp.zeros((1, self.size), dtype=self.range.dtype)
        with jax.ensure_compile_time_eval():
            c = jnp.repeat(
                jnp.eye(self.size, dtype=self.range.dtype),
                np.array(self.shape) - 1,
                axis=0,
            )
            combs = jnp.array(
                list(itertools.combinations(np.arange(len(c)), self.n_particles))
            )
            all_states = c[combs].sum(axis=1, dtype=self.range.dtype)
            if (np.array(self.shape) > 1).any():
                all_states = jnp.unique(all_states, axis=0)
        all_states_fock = jnp.asarray(all_states)
        return self.range.numbers_to_states(all_states_fock, dtype=self.range.dtype)

    @struct.property_cached(pytree_node=True)
    def _lookup_table(self) -> LookupTableHilbertIndex:
        with jax.ensure_compile_time_eval():
            all_states = self._compute_all_states()
        return LookupTableHilbertIndex(all_states)

    @property
    def n_states_bound(self):
        # upper bound on n_states, exact if n_max == 1
        # number of combinations to check in _compute_all_states
        return math.comb((np.array(self.shape) - 1).sum(), self.n_particles)

    @property
    def is_indexable(self):
        # make sure we have less than than max_states to check in _compute_all_states
        return is_indexable(self.n_states_bound)



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import sqrt

import numpy as np

from netket.graph import AbstractGraph, Lattice
from netket.utils import HashableArray
from netket.utils.types import Array


def check_reorder_idx(
    reorder_idx: HashableArray | None,
    inv_reorder_idx: HashableArray | None,
    prev_neighbors: HashableArray | None,
):
    """
    Check that the reordering indices determining the autoregressive order of an
    RNN are correctly declared.
    See :class:`netket.experimental.models.RNN` for details about the reordering indices.
    """
    if reorder_idx is None and inv_reorder_idx is None and prev_neighbors is None:
        # There is a faster code path for 1D RNN
        return

    if reorder_idx is None or inv_reorder_idx is None or prev_neighbors is None:
        raise ValueError(
            "`reorder_idx`, `inv_reorder_idx`, and `prev_neighbors` must be "
            "provided at the same time."
        )

    reorder_idx = np.asarray(reorder_idx)
    inv_reorder_idx = np.asarray(inv_reorder_idx)
    prev_neighbors = np.asarray(prev_neighbors)

    if reorder_idx.ndim != 1:
        raise ValueError("`reorder_idx` must be 1D.")
    if inv_reorder_idx.ndim != 1:
        raise ValueError("`inv_reorder_idx` must be 1D.")
    if prev_neighbors.ndim != 2:
        raise ValueError("`prev_neighbors` must be 2D.")

    V = reorder_idx.size
    if inv_reorder_idx.size != V:
        raise ValueError(
            "`reorder_idx` and `inv_reorder_idx` must have the same length."
        )
    if prev_neighbors.shape[0] != V:
        raise ValueError(
            "`reorder_idx` and `prev_neighbors` must have the same length."
        )

    idx = reorder_idx[inv_reorder_idx]
    if not np.array_equal(idx, np.arange(V)):
        raise ValueError("`inv_reorder_idx` is not the inverse of `reorder_idx`.")

    for i in range(V):
        n = prev_neighbors[i]
        for j in n:
            if j < -1 or j >= V:
                raise ValueError(f"Invaild neighbor {j} of site {i}")

        n = [j for j in n if j >= 0]
        if len(set(n)) != len(n):
            raise ValueError(f"Duplicate neighbor {n} of site {i}")

        for j in n:
            if reorder_idx[j] >= reorder_idx[i]:
                raise ValueError(f"Site {j} is not a previous neighbor of site {i}")


def ensure_prev_neighbors(
    *,
    reorder_idx: HashableArray | None = None,
    inv_reorder_idx: HashableArray | None = None,
    prev_neighbors: HashableArray | None = None,
    graph: AbstractGraph = None,
    check: bool = False,
) -> tuple[HashableArray, HashableArray, HashableArray]:
    """
    Deduce the missing arguments between *reorder_idx*,
    *inv_reorder_idx*, and *inv_reorder_idx* from the specified arguments.

    See :class:`netket.experimental.models.RNN` for details about the reordering indices.

    If no information on neighbors or graph is provided, assumes a 1D ordering.

    Args:
        reorder_idx: indices to transform the inputs from unordered to ordered.
            See :meth:`netket.models.AbstractARNN.reorder` for details
        inv_reorder_idx: indices to transform the inputs from ordered to unordered.
            See :meth:`netket.models.AbstractARNN.reorder` for details.
        prev_neighbors: previous neighbors of each site.
        graph: graph of the physical system, to deduce neighbors.
        check: check the validity of the provided values.
    """
    if inv_reorder_idx is None and graph is not None:
        inv_reorder_idx = _get_inv_reorder_idx(graph)

    if reorder_idx is None:
        if inv_reorder_idx is not None:
            reorder_idx = _get_inv_idx(inv_reorder_idx)
    else:
        if inv_reorder_idx is None:
            inv_reorder_idx = _get_inv_idx(reorder_idx)
    # Now reorder_idx and inv_reorder_idx must be both None or not None

    if reorder_idx is None:
        if prev_neighbors is None:
            # There is a faster code path for 1D RNN
            pass
        else:
            raise ValueError(
                "When `prev_neighbors` is provided, you must also provide "
                "either `reorder_idx` or `inv_reorder_idx`."
            )
    else:
        if prev_neighbors is None:
            if graph is None:
                raise ValueError(
                    "When `reorder_idx` is provided, you must also provide "
                    "either `prev_neighbors` or `graph`."
                )
            else:
                prev_neighbors = _get_prev_neighbors(graph, reorder_idx)

    # Validity of the values will be checked by `_check_reorder_idx` in `RNNLayer`
    if check:
        check_reorder_idx(reorder_idx, inv_reorder_idx, prev_neighbors)

    return reorder_idx, inv_reorder_idx, prev_neighbors


def _get_inv_idx(idx: Array) -> HashableArray:
    idx = np.asarray(idx)
    inv = np.empty_like(idx)
    for i, k in enumerate(idx):
        inv[k] = i
    inv = HashableArray(inv)
    return inv


def _get_extent(graph: AbstractGraph) -> tuple[int, int, int]:
    V = graph.n_nodes

    if isinstance(graph, Lattice):
        assert len(graph.extent) == 2
        L, M = graph.extent
    else:
        L = M = int(sqrt(V))
    assert L * M == V

    return V, L, M


def get_snake_inv_reorder_idx(graph: AbstractGraph) -> HashableArray:
    """
    A helper function to generate the inverse reorder indices in the snake order
    for a 2D graph.

    See :class:`netket.experimental.models.RNN` for details about the reordering indices.
    """
    V, L, M = _get_extent(graph)
    idx = np.arange(V, dtype=np.intp).reshape((L, M))
    idx[1::2, :] = idx[1::2, ::-1]
    idx = idx.flatten()
    idx = HashableArray(idx)
    return idx


def _get_inv_reorder_idx(graph: AbstractGraph) -> HashableArray:
    """
    A greedy algorithm to determine an autoregressive order with good locality.
    For any rectangular graph with OBC, it is the same as the snake order.
    For PBC, it is mostly the same except at the end of each row.

    Start from site 0;
    at each step, choose the unvisited neighbor whose index is the closest to the last one;
    if two neighbors have the same index distance, choose the one with the smaller index;
    if there is no unvisited neighbor, choose the unvisited site with the smallest index.

    Args:
      graph: A :class:`netket.graph.AbstractGraph` instance.

    Returns:
      A hashable array of ints describing the indices to transform an array from ordered to unordered.
      See :meth:`netket.models.AbstractARNN.reorder` for details.
    """
    V = graph.n_nodes
    adj = graph.adjacency_list()
    idx = np.empty(V, dtype=np.intp)
    visited = np.zeros(V, dtype=bool)

    idx[0] = 0
    visited[0] = True
    for i in range(1, V):
        last_k = idx[i - 1]
        neighbors = [x for x in adj[last_k] if not visited[x]]
        if neighbors:
            k = min([(abs(x - last_k), x) for x in neighbors])[1]
        else:
            k = next(x for x in range(V) if not visited[x])
        idx[i] = k
        visited[k] = True

    idx = HashableArray(idx)
    return idx


def _get_snake_prev_neighbors(graph: AbstractGraph) -> HashableArray:
    V, L, M = _get_extent(graph)

    def h(i, j):
        if 0 <= i < L and 0 <= j < M:
            return i * M + j
        else:
            return -1

    def get_neighbors(k):
        i, j = divmod(k, M)
        return h(i, j - 1 if i % 2 == 0 else j + 1), h(i - 1, j)

    # Sort and put -1 padding at the end
    n = [sorted(get_neighbors(k), key=lambda x: (x < 0, x)) for k in range(V)]
    n = np.asarray(n, dtype=np.intp)
    n = HashableArray(n)
    return n


def _get_prev_neighbors(
    graph: AbstractGraph, reorder_idx: Array, max_prev_neighbors=None
) -> HashableArray:
    adj = graph.adjacency_list()
    reorder_idx = np.asarray(reorder_idx)

    n = [[y for y in x if reorder_idx[y] < reorder_idx[i]] for i, x in enumerate(adj)]
    # By default we take the median number of previous neighbors, rounded up
    if max_prev_neighbors is None:
        max_prev_neighbors = int(np.median([len(x) for x in n]) + 0.5)
    # When a site has more previous neighbors, select the ones with largest indices
    n = [x[-max_prev_neighbors:] for x in n]
    n = [sorted(x) + [-1] * (max_prev_neighbors - len(x)) for x in n]

    n = np.asarray(n, dtype=np.intp)
    n = HashableArray(n)
    return n



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

from netket.utils import mpi


def subtract_mean(x, axis=None):
    """
    Subtracts the mean of the input array over all but the last dimension
    and over all MPI processes from each entry.

    Args:
        x: Input array
        axis: Axis or axes along which the means are computed. The default (None) is to
              compute the mean of the flattened array.

    Returns:
        The resulting array.

    """
    # here we keep the dims, since automatic broadcasting of a scalar (shape () )
    # to an array produces errors when used inside of a function which is transposed
    # with jax.linear_transpose
    x_mean = mean(x, axis=axis, keepdims=True)
    return x - x_mean  # automatic broadcasting of x_mean


def mean(a, axis=None, keepdims: bool = False):
    """
    Compute the arithmetic mean along the specified axis and over MPI processes.

    Returns the average of the array elements. The average is taken over the flattened
    array by default, otherwise over the specified axis. float64 intermediate and
    return values are used for integer inputs.

    Args:
        a: The input array
        axis: Axis or axes along which the means are computed. The default (None) is
            to compute the mean of the flattened array.
        keepdims: If True the output array will have the same number of dimensions as
            the input, with the reduced axes having length 1. (default=False)

    Returns:
        The array with reduced dimensions defined by axis.

    """
    out = a.mean(axis=axis, keepdims=keepdims)

    out, _ = mpi.mpi_mean_jax(out)
    return out


def sum(a, axis=None, keepdims: bool = False):
    """
    Compute the sum along the specified axis and over MPI processes.

    Args:
        a: The input array
        axis: Axis or axes along which the mean is computed. The default (None) is to
              compute the mean of the flattened array.
        out: An optional pre-allocated array to fill with the result.
        keepdims: If True the output array will have the same number of dimensions as
              the input, with the reduced axes having length 1. (default=False)

    Returns:
        The array with reduced dimensions defined by axis. If out is not none,
        returns out.

    """
    # if it's a numpy-like array...
    if hasattr(a, "shape"):
        # use jax
        a_sum = a.sum(axis=axis, keepdims=keepdims)
    else:
        # assume it's a scalar
        a_sum = jnp.asarray(a)

    out, _ = mpi.mpi_sum_jax(a_sum)
    return out


def var(a, axis=None, ddof: int = 0):
    """
    Compute the variance mean along the specified axis and over MPI processes.
    Assumes same shape on all MPI processes.

    Args:
        a: The input array
        axis: Axis or axes along which the variance is computed. The default (None)
            is to compute the variance of the whole flattened array.
        out: An optional pre-allocated array to fill with the result.
        ddof: “Delta Degrees of Freedom”: the divisor used in the calculation
            is $N - ddof$, where $N$ represents the number of elements.
            By default ddof is zero.

    Returns:
        The array with reduced dimensions defined by axis. If out is not none,
        returns out.

    """
    m = mean(a, axis=axis)

    if axis is None:
        ssq = jnp.abs(a - m) ** 2.0
    else:
        ssq = jnp.abs(a - jnp.expand_dims(m, axis)) ** 2.0

    out = sum(ssq, axis=axis)

    n_all = total_size(a, axis=axis)
    out /= n_all - ddof

    return out


def total_size(a, axis=None):
    """
    Compute the total number of elements stored in the input array among all
    MPI processes.

    This function essentially returns MPI_sum_among_processes(a.size).

    Args:
        a: The input array.
        axis: If specified, only considers the total size of that axis.

    Returns:
        a.size or a.shape[axis], reduced among all MPI processes.
    """
    if axis is None:
        l_size = a.size
    else:
        l_size = a.shape[axis]

    # TODO: This function cannot call Python MPI because if it gets called on shape
    # inference when compiling. Therefore if only one mpi rank is compiling this
    # leads to deadlocks.
    # We should refactor all this logic.
    # return mpi_sum(l_size)
    return l_size * mpi.n_nodes



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import jax
import jax.numpy as jnp

from flax import linen as nn
from jax.nn.initializers import lecun_normal, zeros

from netket.utils.types import NNInitFunc, DType
from netket import nn as nknn

default_kernel_init = lecun_normal()
default_bias_init = zeros


class MLP(nn.Module):
    r"""A Multi-Layer Perceptron with hidden layers.

    This model uses the MLP block with output dimension 1, which is squeezed.

    This combines multiple dense layers and activations functions into a single object.
    It separates the output layer from the hidden layers,
    since it typically has a different form.
    One can specify the specific activation functions per layer.
    The size of the hidden dimensions can be provided as a number,
    or as a factor relative to the input size (similar as for RBM).
    The default model is a single linear layer without activations.

    Forms a common building block for models such as
    `PauliNet (continuous) <https://www.nature.com/articles/s41557-020-0544-y>`_
    """

    hidden_dims: int | tuple[int, ...] | None = None
    """The size of the hidden layers, excluding the output layer."""
    hidden_dims_alpha: int | tuple[int, ...] | None = None
    """The size of the hidden layers provided as number of times the input size.
    One must choose to either specify this or the hidden_dims keyword argument"""
    param_dtype: DType = jnp.float64
    """The dtype of the weights."""
    hidden_activations: Callable | tuple[Callable, ...] | None = nknn.gelu
    """The nonlinear activation function after each hidden layer.
    Can be provided as a single activation,
    where the same activation will be used for every layer."""
    output_activation: Callable | None = None
    """The nonlinear activation at the output layer.
    If None is provided, the output layer will be essentially linear."""
    use_hidden_bias: bool = True
    """if True uses a bias in the hidden layer."""
    use_output_bias: bool = False
    """if True adds a bias to the output layer."""
    precision: jax.lax.Precision | None = None
    """Numerical precision of the computation see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the Dense layer matrix."""
    bias_init: NNInitFunc = default_bias_init
    """Initializer for the biases."""

    @nn.compact
    def __call__(self, input):
        x = nknn.blocks.MLP(
            output_dim=1,  # a netket model has a single output
            hidden_dims=self.hidden_dims,
            hidden_dims_alpha=self.hidden_dims_alpha,
            param_dtype=self.param_dtype,
            hidden_activations=self.hidden_activations,
            output_activation=self.output_activation,
            use_hidden_bias=self.use_hidden_bias,
            use_output_bias=self.use_output_bias,
            precision=self.precision,
            kernel_init=self.kernel_init,
            bias_init=self.bias_init,
        )(input)
        x = x.squeeze(-1)
        return x



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, lru_cache
from collections.abc import Callable

import jax
from jax import numpy as jnp
from flax.core.scope import CollectionFilter, DenyList  # noqa: F401

from netket import jax as nkjax
from netket.operator import DiscreteOperator, Squared
from netket.stats import Stats
from netket.utils.types import PyTree
from netket.utils.dispatch import dispatch

from netket.vqs import expect_and_grad, expect_and_forces
from netket.vqs.mc.common import force_to_grad

from .state import FullSumState


def _check_hilbert(A, B):
    if A.hilbert != B.hilbert:
        raise NotImplementedError(  # pragma: no cover
            f"Non matching hilbert spaces {A.hilbert} and {B.hilbert}"
        )


# TODO: This cache is here so that we don't re-compute the sparse representation of the operators at every VMC step
# but instead we cache the last 5 used. Should investigate a better way to implement this caching.
@lru_cache(5)
def sparsify(Ô):
    """
    Converts to sparse but also cache the sparsificated result to speed up.
    """
    return Ô.to_sparse()


@dispatch
def expect(vstate: FullSumState, Ô: DiscreteOperator) -> Stats:  # noqa: F811
    _check_hilbert(vstate, Ô)

    O = sparsify(Ô)
    Ψ = vstate.to_array()

    # TODO: This performs the full computation on all MPI ranks.
    # It would be great if we could split the computation among ranks.

    OΨ = O @ Ψ
    expval_O = (Ψ.conj() * OΨ).sum()

    variance = jnp.sum(jnp.abs(OΨ - expval_O * Ψ) ** 2)
    return Stats(mean=expval_O, error_of_mean=0.0, variance=variance)


@expect_and_grad.dispatch
def expect_and_grad_fullsum(
    vstate: FullSumState,
    Ô: DiscreteOperator,
    *args,
    mutable: CollectionFilter = False,
    use_covariance: bool | None = None,
    **kwargs,
) -> tuple[Stats, PyTree]:
    if use_covariance is None:
        use_covariance = Ô.is_hermitian

    if use_covariance:
        # Implementation of expect_and_grad for `use_covariance == True` (due to the Literal[True]
        # type in the signature).` This case is equivalent to the composition of the
        # `expect_and_forces` and `force_to_grad` functions.
        # return expect_and_grad_from_covariance(vstate, Ô, *args, mutable=mutable)
        Ō, Ō_grad = expect_and_forces(vstate, Ô, *args, mutable=mutable, **kwargs)
        Ō_grad = force_to_grad(Ō_grad, vstate.parameters)
        return Ō, Ō_grad
    else:
        raise NotImplementedError(
            "The gradient of non hermitian operators with "
            "FullSumState is not yet implemented"
        )


@expect_and_forces.dispatch
def expect_and_forces_fullsum(
    vstate: FullSumState,
    Ô: DiscreteOperator,
    *,
    mutable: CollectionFilter = False,
) -> tuple[Stats, PyTree]:
    if isinstance(Ô, Squared):
        raise NotImplementedError("expect_and_forces not yet implemented for `Squared`")

    _check_hilbert(vstate, Ô)

    O = sparsify(Ô)
    Ψ = vstate.to_array()
    OΨ = O @ Ψ

    expval_O, Ō_grad, new_model_state = _exp_forces(
        vstate._apply_fun,
        mutable,
        vstate.parameters,
        vstate.model_state,
        vstate._all_states,
        OΨ,
        Ψ,
    )

    if mutable is not False:
        vstate.model_state = new_model_state

    return expval_O, Ō_grad


@partial(jax.jit, static_argnums=(0, 1))
def _exp_forces(
    model_apply_fun: Callable,
    mutable: CollectionFilter,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    OΨ: jnp.ndarray,
    Ψ: jnp.ndarray,
) -> tuple[PyTree, PyTree]:
    is_mutable = mutable is not False

    expval_O = (Ψ.conj() * OΨ).sum()
    variance = jnp.sum(jnp.abs(OΨ - expval_O * Ψ) ** 2)

    ΔOΨ = (OΨ - expval_O * Ψ).conj() * Ψ

    _, vjp_fun, *new_model_state = nkjax.vjp(
        lambda w: model_apply_fun({"params": w, **model_state}, σ, mutable=mutable),
        parameters,
        conjugate=True,
        has_aux=is_mutable,
    )

    Ō_grad = vjp_fun(ΔOΨ)[0]

    new_model_state = new_model_state[0] if is_mutable else None

    return (
        Stats(mean=expval_O, error_of_mean=0.0, variance=variance),
        Ō_grad,
        new_model_state,
    )



import math

import jax
import jax.numpy as jnp

from netket.utils import struct
from netket.utils.types import Array

from netket.hilbert.constraint import SumOnPartitionConstraint

from .base import HilbertIndex, is_indexable
from .unconstrained import LookupTableHilbertIndex
from .constrained_sum import SumConstrainedHilbertIndex
from .constrained_generic import optimalConstrainedHilbertindex


@optimalConstrainedHilbertindex.dispatch
def optimalConstrainedHilbertindex(
    local_states, size, constraint: SumOnPartitionConstraint
):
    if size != sum(constraint.sizes):
        raise ValueError(
            f"The size {size} does not match the total size of the constraint"
            f"partitions {sum(constraint.sizes)}."
        )
    specialized_index = SumOnPartitionConstrainedHilbertIndex(
        sub_indices=tuple(
            SumConstrainedHilbertIndex(local_states, N, sv)
            for (N, sv) in zip(constraint.sizes, constraint.sum_values)
        )
    )
    return specialized_index


@struct.dataclass
class SumOnPartitionConstrainedHilbertIndex(HilbertIndex):
    """
    Specialized implementation for a constrained space with a SumConstraint.
    Does not require the unconstrained space to be indexable.
    """

    sub_indices: list[SumConstrainedHilbertIndex] = struct.field(pytree_node=False)

    @property
    def size(self) -> int:
        return sum(s.size for s in self.sub_indices)

    @property
    def shape(self):
        return sum((s.shape for s in self.sub_indices), start=())

    @property
    def n_states(self):
        return math.prod(s.n_states for s in self.sub_indices)

    def states_to_numbers(self, states: Array) -> Array:
        return self._lookup_table.states_to_numbers(states)

    def numbers_to_states(self, numbers: Array):
        return self._lookup_table.numbers_to_states(numbers)

    def all_states(self):
        return self._lookup_table.all_states()

    @jax.jit
    def _compute_all_states(self):
        states = []
        for index in self.sub_indices:
            states.append(index.all_states())
        return combine_configurations(*states)

    @struct.property_cached(pytree_node=True)
    def _lookup_table(self) -> LookupTableHilbertIndex:
        with jax.ensure_compile_time_eval():
            all_states = self._compute_all_states()
        return LookupTableHilbertIndex(all_states)

    @property
    def n_states_bound(self):
        # upper bound on n_states, exact if n_max == 1
        # number of combinations to check in _compute_all_states
        return math.prod(s.n_states_bound for s in self.sub_indices)

    @property
    def is_indexable(self):
        # make sure we have less than than max_states to check in _compute_all_states
        return is_indexable(self.n_states_bound)


def combine_configurations(*matrices):
    """
    Concatenate a list of configurations, generating all possible combinations.
    """
    if len(matrices) == 1:
        return matrices[0]

    # Initialize with the first matrix
    result = matrices[0]

    for matrix in matrices[1:]:
        M1, N1 = result.shape
        M2, N2 = matrix.shape

        # Reshape matrices to make use of broadcasting
        result_reshaped = result[:, jnp.newaxis, :]
        matrix_reshaped = matrix[jnp.newaxis, :, :]

        # Create the resulting matrix by concatenating along the last axis
        result = jnp.concatenate(
            (result_reshaped.repeat(M2, axis=1), matrix_reshaped.repeat(M1, axis=0)),
            axis=-1,
        )

        # Reshape to the new form
        result = result.reshape(M1 * M2, N1 + N2)

    return result



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import numpy as np

from netket.utils import struct


class ConvergenceStopping(struct.Pytree, mutable=True):
    """A simple callback to stop the optimisation when the monitored quantity gets
    below a certain threshold for at least `patience` steps.
    """

    target: float = struct.field(serialize=False)
    """Target value for the monitored quantity. Training will stop if the driver drops below this value."""
    monitor: str = struct.field(serialize=False)
    """Loss statistic to monitor. Should be one of 'mean', 'variance', 'error_of_mean'."""
    smoothing_window: int = struct.field(serialize=False)
    """The loss is smoothed over the last `smoothing_window` iterations to
    reduce statistical fluctuations"""
    patience: int = struct.field(serialize=False)
    """The loss must be consistently below this value for this number of
    iterations in order to stop the optimisation."""

    # caches
    _loss_window: deque
    _patience_counter: int

    def __init__(
        self,
        target: float,
        monitor: str = "mean",
        *,
        smoothing_window: int = 10,
        patience: int = 10,
    ):
        """
        Construct a callback stopping the optimisation when the monitored quantity
        gets below a certain threshold for at least `patience` steps.

        Args:
            target: the threshold value for the monitored quantity. Training will stop if the driver drops below this value.
            monitor: a string with the name of the quantity to be monitored. This
                is applied to the standard loss optimised by a driver, such as the
                Energy for the VMC driver. Should be one of
                'mean', 'variance', 'error_of_mean' (default: 'mean').
            smoothing_window: an integer number of steps over which the monitored value
                is averaged before comparing to target.
            patience: Number of steps to wait before stopping the execution after
                the tracked quantity drops below the target value (default 0, meaning
                that it stops immediately).
        """
        self.target = target
        self.monitor = monitor
        self.smoothing_window = smoothing_window
        self.patience = patience

        self._loss_window: deque = deque([], maxlen=self.smoothing_window)
        self._patience_counter: int = 0

    def __call__(self, step, log_data, driver):
        """
        A boolean function that determines whether or not to stop training.

        Args:
            step: An integer corresponding to the step (iteration or epoch) in training.
            log_data: A dictionary containing log data for training.
            driver: A NetKet variational driver.

        Returns:
            A boolean. If True, training continues, else, it does not.
        """
        loss = np.asarray(np.real(getattr(log_data[driver._loss_name], self.monitor)))

        self._loss_window.append(loss)
        loss_smooth = np.mean(self._loss_window)

        if loss_smooth <= self.target:
            self._patience_counter += 1
        else:
            self._patience_counter = 0

        if self._patience_counter > self.patience:
            return False

        return True



from types import ModuleType

import importlib


def version_tuple(verstr: str):
    # drop everything after the numeric part of the version
    allowed_chars = "0123456789."
    for i, char in enumerate(verstr):
        if char not in allowed_chars:
            break
    else:
        i = len(verstr) + 1

    verstr = verstr[:i].rstrip(".")
    vertupl = tuple(int(v) for v in verstr.split("."))[:3]

    # Ensure that we have (major, minor, patch)
    while len(vertupl) < 3:
        vertupl = vertupl + (0,)
    return vertupl


def module_version(module: str | ModuleType) -> tuple[int, ...]:
    if isinstance(module, str):
        module = importlib.import_module(module)

    return version_tuple(module.__version__)


def version_string(module: str | ModuleType) -> str:
    if isinstance(module, str):
        module = importlib.import_module(module)

    return module.__version__



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any
from collections.abc import Callable
from textwrap import dedent

import numpy as np

import jax
from flax import linen as nn
from jax import numpy as jnp

from netket.hilbert import AbstractHilbert, ContinuousHilbert

from netket.utils import mpi, wrap_afun
from netket.utils.types import PyTree, DType

from netket.utils.deprecation import warn_deprecation
from netket.utils import struct

from netket.jax.sharding import (
    device_count,
    shard_along_axis,
)
from netket.jax import apply_chunked, dtype_real

from .base import Sampler, SamplerState
from .rules import MetropolisRule


class MetropolisSamplerState(SamplerState):
    """
    State for a Metropolis sampler.

    Contains the current configuration, the RNG state and the (optional)
    state of the transition rule.
    """

    σ: jnp.ndarray = struct.field(
        sharded=struct.ShardedFieldSpec(
            sharded=True, deserialization_function="relaxed-ignore-errors"
        )
    )
    """Current batch of configurations in the Markov chain."""
    log_prob: jnp.ndarray = struct.field(sharded=True, serialize=False)
    """Log probabilities of the current batch of configurations σ in the Markov chain."""
    rng: jnp.ndarray = struct.field(
        sharded=struct.ShardedFieldSpec(
            sharded=True, deserialization_function="relaxed-rng-key"
        )
    )
    """State of the random number generator (key, in jax terms)."""
    rule_state: Any | None
    """Optional state of the transition rule."""

    n_steps_proc: int = struct.field(default_factory=lambda: jnp.zeros((), dtype=int))
    """Number of moves performed along the chains in this process since the last reset."""
    n_accepted_proc: jnp.ndarray = struct.field(
        sharded=struct.ShardedFieldSpec(
            sharded=True, deserialization_function="relaxed-ignore-errors"
        )
    )
    """Number of accepted transitions among the chains in this process since the last reset."""

    def __init__(
        self,
        σ: jnp.ndarray,
        rng: jnp.ndarray,
        rule_state: Any | None,
        log_prob: jnp.ndarray | None = None,
    ):
        self.σ = σ
        self.rng = rng
        self.rule_state = rule_state

        if log_prob is None:
            log_prob = jnp.full(self.σ.shape[:-1], -jnp.inf, dtype=float)
        self.log_prob = shard_along_axis(log_prob, axis=0)

        self.n_accepted_proc = shard_along_axis(
            jnp.zeros(σ.shape[0], dtype=int), axis=0
        )
        self.n_steps_proc = jnp.zeros((), dtype=int)
        super().__init__()

    @property
    def acceptance(self) -> float:
        """The fraction of accepted moves across all chains and MPI processes.

        The rate is computed since the last reset of the sampler.
        Will return None if no sampling has been performed since then.
        """
        if self.n_steps == 0:
            return None

        return self.n_accepted / self.n_steps

    @property
    def n_steps(self) -> int:
        """Total number of moves performed across all processes since the last reset."""
        return self.n_steps_proc * mpi.n_nodes

    @property
    def n_accepted(self) -> int:
        """Total number of moves accepted across all processes since the last reset."""
        # jit sum for gda
        res, _ = mpi.mpi_sum_jax(jax.jit(jnp.sum)(self.n_accepted_proc))
        return res

    def __repr__(self):
        if self.n_steps > 0:
            acc_string = f"# accepted = {self.n_accepted}/{self.n_steps} ({self.acceptance * 100}%), "
        else:
            acc_string = ""

        return f"{type(self).__name__}({acc_string}rng state={self.rng})"


def _assert_good_sample_shape(samples, shape, dtype, obj=""):
    canonical_dtype = jax.dtypes.canonicalize_dtype(dtype)
    if samples.shape != shape or samples.dtype != canonical_dtype:
        raise ValueError(
            dedent(
                f"""

            The samples returned by the {obj} have `shape={samples.shape}` and
            `dtype={samples.dtype}`, but the sampler requires `shape={shape} and
            `dtype={canonical_dtype}` (canonicalized from {dtype}).

            If you are using a custom transition rule, check that it returns the
            correct shape and dtype.

            If you are using a built-in transition rule, there might be a mismatch
            between hilbert spaces, or it's a bug in NetKet.

            """
            )
        )


def _assert_good_log_prob_shape(log_prob, n_chains_per_rank, machine):
    if log_prob.shape != (n_chains_per_rank,):
        raise ValueError(
            dedent(
                f"""

            The output of the model {machine} has `shape={log_prob.shape}`, but
            `shape=({n_chains_per_rank},)` was expected.

            This might be because of an hilbert space mismatch or because your
            model is ill-configured.

            """
            )
        )


def _round_n_chains_to_next_multiple(
    n_chains, n_chains_per_whatever, n_whatever, whatever_str
):
    # small helper function to round the number of chains to the next multiple of [whatever]
    # here [whatever] can be e.g. mpi ranks or jax devices
    # if n_chains is None and n_chains_per_whatever is None:
    #    n_chains_per_whatever = default
    if n_chains is not None and n_chains_per_whatever is not None:
        raise ValueError(
            f"Cannot specify both `n_chains` and `n_chains_per_{whatever_str}`"
        )
    elif n_chains is not None:
        n_chains_per_whatever = max(int(np.ceil(n_chains / n_whatever)), 1)
        if n_chains_per_whatever * n_whatever != n_chains:
            if mpi.rank == 0:
                import warnings

                warnings.warn(
                    f"Using {n_chains_per_whatever} chains per {whatever_str} among {n_whatever} {whatever_str}s "
                    f"(total={n_chains_per_whatever * n_whatever} instead of n_chains={n_chains}). "
                    f"To directly control the number of chains on every {whatever_str}, specify "
                    f"`n_chains_per_{whatever_str}` when constructing the sampler. "
                    f"To silence this warning, either use `n_chains_per_{whatever_str}` or use `n_chains` "
                    f"that is a multiple of the number of {whatever_str}s",
                    category=UserWarning,
                    stacklevel=2,
                )
    return n_chains_per_whatever * n_whatever


class MetropolisSampler(Sampler):
    r"""
    Metropolis-Hastings sampler for a Hilbert space according to a specific transition rule.

    The transition rule is used to generate a proposed state :math:`s^\prime`, starting from the
    current state :math:`s`. The move is accepted with probability

    .. math::

        A(s \rightarrow s^\prime) = \mathrm{min} \left( 1,\frac{P(s^\prime)}{P(s)} e^{L(s,s^\prime)} \right) ,

    where the probability being sampled from is :math:`P(s)=|M(s)|^p`. Here :math:`M(s)` is a
    user-provided function (the machine), :math:`p` is also user-provided with default value :math:`p=2`,
    and :math:`L(s,s^\prime)` is a suitable correcting factor computed by the transition kernel.

    The dtype of the sampled states can be chosen.
    """

    rule: MetropolisRule = None
    """The Metropolis transition rule."""
    sweep_size: int = struct.field(pytree_node=False, default=None)
    """Number of sweeps for each step along the chain. Defaults to the number
    of sites in the Hilbert space."""
    n_chains: int = struct.field(pytree_node=False)
    """Total number of independent chains across all MPI ranks and/or devices."""
    chunk_size: int | None = struct.field(pytree_node=False, default=None)
    """Chunk size for evaluating wave functions."""
    reset_chains: bool = struct.field(pytree_node=False, default=False)
    """If True, resets the chain state when `reset` is called on every new sampling."""

    def __init__(
        self,
        hilbert: AbstractHilbert,
        rule: MetropolisRule,
        *,
        n_sweeps: int = None,
        sweep_size: int = None,
        reset_chains: bool = False,
        n_chains: int | None = None,
        n_chains_per_rank: int | None = None,
        chunk_size: int | None = None,
        machine_pow: int = 2,
        dtype: DType = float,
    ):
        """
        Constructs a Metropolis Sampler.

        Args:
            hilbert: The Hilbert space to sample.
            rule: A `MetropolisRule` to generate random transitions from a given state as
                well as uniform random states.
            n_chains: The total number of independent Markov chains across all MPI ranks.
                Either specify this or `n_chains_per_rank`. If MPI is disabled, the two are equivalent;
                if MPI is enabled and `n_chains` is specified, then every MPI rank will run
                `n_chains/mpi.n_nodes` chains. In general, we recommend specifying `n_chains_per_rank`
                as it is more portable.
            n_chains_per_rank: Number of independent chains on every MPI rank (default = 16).
                                If netket_experimental_sharding is enabled this is interpreted as the number
                                of independent chains on every jax device, and the n_chains_per_rank
                                property of the sampler will return the total number of chains on all devices.
            chunk_size: Chunk size for evaluating the ansatz while sampling. Must divide n_chains_per_rank.
            sweep_size: Number of sweeps for each step along the chain.
                This is equivalent to subsampling the Markov chain. (Defaults to the number of sites
                in the Hilbert space.)
            reset_chains: If True, resets the chain state when `reset` is called on every
                new sampling (default = False).
            machine_pow: The power to which the machine should be exponentiated to generate
                the pdf (default = 2).
            dtype: The dtype of the states sampled (default = np.float64).
        """

        # Validate the inputs
        if not isinstance(rule, MetropolisRule):
            raise TypeError(
                f"The second positional argument, rule, must be a MetropolisRule but "
                f"`type(rule)={type(rule)}`."
            )

        if not isinstance(reset_chains, bool):
            raise TypeError("reset_chains must be a boolean.")

        if n_sweeps is not None:
            warn_deprecation(
                "Specifying `n_sweeps` when constructing sampler is deprecated. Please use `sweep_size` instead."
            )
            if sweep_size is not None:
                raise ValueError("Cannot specify both `sweep_size` and `n_sweeps`")
            sweep_size = n_sweeps

        if sweep_size is None:
            sweep_size = hilbert.size

        # Default n_chains per rank, if unset
        if n_chains is None and n_chains_per_rank is None:
            # TODO set it to a few hundred if on GPU?
            n_chains_per_rank = 16

        n_chains = _round_n_chains_to_next_multiple(
            n_chains,
            n_chains_per_rank,
            device_count(),
            "rank",
        )
        n_chains_per_rank = n_chains // device_count()

        if chunk_size is not None and n_chains_per_rank % chunk_size != 0:
            raise ValueError(
                f"Chunk size must divide number of chains per rank, {n_chains_per_rank}"
            )
        self.chunk_size = chunk_size

        super().__init__(
            hilbert=hilbert,
            machine_pow=machine_pow,
            dtype=dtype,
        )

        self.n_chains = n_chains
        self.reset_chains = reset_chains
        self.rule = rule
        self.sweep_size = sweep_size

    @property
    def n_sweeps(self):
        warn_deprecation(
            "`MetropolisSampler.n_sweeps` is deprecated. Please use `MetropolisSampler.sweep_size` instead."
        )
        return self.sweep_size

    def sample_next(
        sampler,
        machine: Callable | nn.Module,
        parameters: PyTree,
        state: SamplerState | None = None,
    ) -> tuple[SamplerState, jnp.ndarray]:
        """
        Samples the next state in the Markov chain.

        Args:
            machine: A Flax module or callable with the forward pass of the log-pdf.
                If it is a callable, it should have the signature
                :code:`f(parameters, σ) -> jnp.ndarray`.
            parameters: The PyTree of parameters of the model.
            state: The current state of the sampler. If not specified, then initialize and reset it.

        Returns:
            state: The new state of the sampler.
            σ: The next batch of samples.

        Note:
            The return order is inverted wrt `sample` because when called inside of
            a scan function the first returned argument should be the state.
        """
        if state is None:
            state = sampler.reset(machine, parameters)

        return sampler._sample_next(wrap_afun(machine), parameters, state)

    @partial(jax.jit, static_argnums=1)
    def _init_state(sampler, machine, parameters, key):
        key_state, key_rule = jax.random.split(key)
        rule_state = sampler.rule.init_state(sampler, machine, parameters, key_rule)
        σ = jnp.zeros((sampler.n_batches, sampler.hilbert.size), dtype=sampler.dtype)
        σ = shard_along_axis(σ, axis=0)

        output_dtype = jax.eval_shape(machine.apply, parameters, σ).dtype
        log_prob = jnp.full(
            (sampler.n_batches,), -jnp.inf, dtype=dtype_real(output_dtype)
        )
        log_prob = shard_along_axis(log_prob, axis=0)

        state = MetropolisSamplerState(
            σ=σ, rng=key_state, rule_state=rule_state, log_prob=log_prob
        )
        # If we don't reset the chain at every sampling iteration, then reset it
        # now.
        if not sampler.reset_chains:
            key_state, rng = jax.jit(jax.random.split)(key_state)
            σ = sampler.rule.random_state(sampler, machine, parameters, state, rng)
            _assert_good_sample_shape(
                σ,
                (sampler.n_batches, sampler.hilbert.size),
                sampler.dtype,
                f"{sampler.rule}.random_state",
            )
            σ = shard_along_axis(σ, axis=0)
            state = state.replace(σ=σ, rng=key_state)
        return state

    @partial(jax.jit, static_argnums=1)
    def _reset(sampler, machine, parameters, state):
        rng = state.rng

        if sampler.reset_chains:
            rng, key = jax.random.split(state.rng)
            σ = sampler.rule.random_state(sampler, machine, parameters, state, rng)
            _assert_good_sample_shape(
                σ,
                (sampler.n_batches, sampler.hilbert.size),
                sampler.dtype,
                f"{sampler.rule}.random_state",
            )
            σ = shard_along_axis(σ, axis=0)
        else:
            σ = state.σ

        # Recompute the log_probability of the current samples
        apply_machine = apply_chunked(
            machine.apply, in_axes=(None, 0), chunk_size=sampler.chunk_size
        )
        log_prob_σ = sampler.machine_pow * apply_machine(parameters, σ).real

        rule_state = sampler.rule.reset(sampler, machine, parameters, state)

        return state.replace(
            σ=σ,
            log_prob=log_prob_σ,
            rng=rng,
            rule_state=rule_state,
            n_steps_proc=jnp.zeros_like(state.n_steps_proc),
            n_accepted_proc=jnp.zeros_like(state.n_accepted_proc),
        )

    def _sample_next(sampler, machine, parameters, state):
        """
        Implementation of `sample_next` for subclasses of `MetropolisSampler`.

        If you subclass `MetropolisSampler`, you should override this and not `sample_next`
        itself, because `sample_next` contains some common logic.
        """
        apply_machine = apply_chunked(
            machine.apply, in_axes=(None, 0), chunk_size=sampler.chunk_size
        )

        def loop_body(i, s):
            # 1 to propagate for next iteration, 1 for uniform rng and n_chains for transition kernel
            s["key"], key1, key2 = jax.random.split(s["key"], 3)

            σp, log_prob_correction = sampler.rule.transition(
                sampler, machine, parameters, state, key1, s["σ"]
            )
            _assert_good_sample_shape(
                σp,
                (sampler.n_batches, sampler.hilbert.size),
                sampler.dtype,
                f"{sampler.rule}.transition",
            )
            proposal_log_prob = sampler.machine_pow * apply_machine(parameters, σp).real
            _assert_good_log_prob_shape(proposal_log_prob, sampler.n_batches, machine)

            uniform = jax.random.uniform(key2, shape=(sampler.n_batches,))
            if log_prob_correction is not None:
                do_accept = uniform < jnp.exp(
                    proposal_log_prob - s["log_prob"] + log_prob_correction
                )
            else:
                do_accept = uniform < jnp.exp(proposal_log_prob - s["log_prob"])

            # do_accept must match ndim of proposal and state (which is 2)
            s["σ"] = jnp.where(do_accept.reshape(-1, 1), σp, s["σ"])
            s["accepted"] += do_accept

            s["log_prob"] = jax.numpy.where(
                do_accept.reshape(-1), proposal_log_prob, s["log_prob"]
            )

            return s

        s = {
            "key": state.rng,
            "σ": state.σ,
            # Log prob is already computed in reset, so don't recompute it.
            # "log_prob": sampler.machine_pow * apply_machine(parameters, state.σ).real,
            "log_prob": state.log_prob,
            # for logging
            "accepted": state.n_accepted_proc,
        }
        s = jax.lax.fori_loop(0, sampler.sweep_size, loop_body, s)

        new_state = state.replace(
            rng=s["key"],
            σ=s["σ"],
            log_prob=s["log_prob"],
            n_accepted_proc=s["accepted"],
            n_steps_proc=state.n_steps_proc + sampler.sweep_size * sampler.n_batches,
        )

        return new_state, new_state.σ

    @partial(jax.jit, static_argnums=(1, 4))
    def _sample_chain(sampler, machine, parameters, state, chain_length):
        """
        Samples `chain_length` batches of samples along the chains.

        Internal method used for jitting calls.

        Arguments:
            sampler: The Monte Carlo sampler.
            machine: A Flax module with the forward pass of the log-pdf.
            parameters: The PyTree of parameters of the model.
            state: The current state of the sampler.
            chain_length: The length of the chains.

        Returns:
            σ: The next batch of samples.
            state: The new state of the sampler
        """
        state, samples = jax.lax.scan(
            lambda state, _: sampler.sample_next(machine, parameters, state),
            state,
            xs=None,
            length=chain_length,
        )
        # make it (n_chains, n_samples_per_chain) as expected by netket.stats.statistics
        samples = jnp.swapaxes(samples, 0, 1)
        return samples, state

    def __repr__(sampler):
        return (
            f"{type(sampler).__name__}("
            + f"\n  hilbert = {sampler.hilbert},"
            + f"\n  rule = {sampler.rule},"
            + f"\n  n_chains = {sampler.n_chains},"
            + f"\n  sweep_size = {sampler.sweep_size},"
            + f"\n  reset_chains = {sampler.reset_chains},"
            + f"\n  machine_power = {sampler.machine_pow},"
            + f"\n  dtype = {sampler.dtype}"
            + ")"
        )

    def __str__(sampler):
        return (
            f"{type(sampler).__name__}("
            + f"rule = {sampler.rule}, "
            + f"n_chains = {sampler.n_chains}, "
            + f"sweep_size = {sampler.sweep_size}, "
            + f"reset_chains = {sampler.reset_chains}, "
            + f"machine_power = {sampler.machine_pow}, "
            + f"dtype = {sampler.dtype})"
        )


def MetropolisLocal(hilbert, **kwargs) -> MetropolisSampler:
    r"""
    Sampler acting on one local degree of freedom.

    This sampler acts locally only on one local degree of freedom :math:`s_i`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s_N`,
    where :math:`s^\prime_i \neq s_i`.

    The transition probability associated to this
    sampler can be decomposed into two steps:

    1. One of the site indices :math:`i = 1\dots N` is chosen with uniform probability.

    2. Among all the possible (:math:`m - 1`) values that :math:`s^\prime_i` can take,
    one of them is chosen with uniform probability.

    For example, in the case of spin :math:`1/2` particles, :math:`m=2`
    and the possible local values are :math:`s_i = -1,+1`.
    In this case then :class:`MetropolisLocal` is equivalent to flipping a random spin.

    In the case of bosons, with occupation numbers
    :math:`s_i = 0, 1, \dots n_{\mathrm{max}}`, :class:`MetropolisLocal`
    would pick a random local occupation number uniformly between :math:`0`
    and :math:`n_{\mathrm{max}}` except the current :math:`s_i`.

    Args:
        hilbert: The Hilbert space to sample.
        n_chains: The total number of independent Markov chains across all MPI ranks. Either specify this or `n_chains_per_rank`.
        n_chains_per_rank: Number of independent chains on every MPI rank (default = 16).
        sweep_size: Number of sweeps for each step along the chain. Defaults to the number of sites in the Hilbert space.
                This is equivalent to subsampling the Markov chain.
        reset_chains: If True, resets the chain state when `reset` is called on every new sampling (default = False).
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the states sampled (default = np.float64).
    """
    from .rules import LocalRule

    return MetropolisSampler(hilbert, LocalRule(), **kwargs)


def MetropolisExchange(
    hilbert, *, clusters=None, graph=None, d_max=1, **kwargs
) -> MetropolisSampler:
    r"""
    This sampler acts locally only on two local degree of freedom :math:`s_i` and :math:`s_j`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s^\prime_j \dots s_N`,
    where in general :math:`s^\prime_i \neq s_i` and :math:`s^\prime_j \neq s_j`.
    The sites :math:`i` and :math:`j` are also chosen to be within a maximum graph
    distance of :math:`d_{\mathrm{max}}`.

    The transition probability associated to this sampler can
    be decomposed into two steps:

    1. A pair of indices :math:`i,j = 1\dots N`, and such that
       :math:`\mathrm{dist}(i,j) \leq d_{\mathrm{max}}`,
       is chosen with uniform probability.

    2. The sites are exchanged, i.e. :math:`s^\prime_i = s_j` and :math:`s^\prime_j = s_i`.

    Notice that this sampling method generates random permutations of the quantum
    numbers, thus global quantities such as the sum of the local quantum numbers
    are conserved during the sampling.
    This scheme should be used then only when sampling in a
    region where :math:`\sum_i s_i = \mathrm{constant}` is needed,
    otherwise the sampling would be strongly not ergodic.

    Args:
        hilbert: The Hilbert space to sample.
        d_max: The maximum graph distance allowed for exchanges.
        n_chains: The total number of independent Markov chains across all MPI ranks. Either specify this or `n_chains_per_rank`.
        n_chains_per_rank: Number of independent chains on every MPI rank (default = 16).
        sweep_size: Number of sweeps for each step along the chain. Defaults to the number of sites in the Hilbert space.
                This is equivalent to subsampling the Markov chain.
        reset_chains: If True, resets the chain state when `reset` is called on every new sampling (default = False).
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the states sampled (default = np.float64).

    Examples:
          Sampling from a RBM machine in a 1D lattice of spin 1/2, using
          nearest-neighbor exchanges.

          >>> import netket as nk
          >>>
          >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
          >>> hi=nk.hilbert.Spin(s=0.5, N=g.n_nodes)
          >>>
          >>> # Construct a MetropolisExchange Sampler
          >>> sa = nk.sampler.MetropolisExchange(hi, graph=g)
          >>> print(sa)
          MetropolisSampler(rule = ExchangeRule(# of clusters: 200), n_chains = 16, sweep_size = 100, reset_chains = False, machine_power = 2, dtype = <class 'float'>)
    """
    from .rules import ExchangeRule

    rule = ExchangeRule(clusters=clusters, graph=graph, d_max=d_max)
    return MetropolisSampler(hilbert, rule, **kwargs)


def MetropolisHamiltonian(hilbert, hamiltonian, **kwargs) -> MetropolisSampler:
    r"""
    Sampling based on the off-diagonal elements of a Hamiltonian (or a generic Operator).
    In this case, the transition matrix is taken to be:

    .. math::
       T( \mathbf{s} \rightarrow \mathbf{s}^\prime) = \frac{1}{\mathcal{N}(\mathbf{s})}\theta(|H_{\mathbf{s},\mathbf{s}^\prime}|),

    where :math:`\theta(x)` is the Heaviside step function, and :math:`\mathcal{N}(\mathbf{s})`
    is a state-dependent normalization.
    The effect of this transition probability is then to connect (with uniform probability)
    a given state :math:`\mathbf{s}` to all those states :math:`\mathbf{s}^\prime` for which the Hamiltonian has
    finite matrix elements.
    Notice that this sampler preserves by construction all the symmetries
    of the Hamiltonian. This is in generally not true for the local samplers instead.

    This sampler only works on the CPU. To use the Hamiltonian sampler with GPUs,
    you should use :class:`netket.sampler.MetropolisHamiltonianNumpy`

    Args:
        hilbert: The Hilbert space to sample.
        hamiltonian: The operator used to perform off-diagonal transition.
        n_chains: The total number of independent Markov chains across all MPI ranks. Either specify this or `n_chains_per_rank`.
        n_chains_per_rank: Number of independent chains on every MPI rank (default = 16).
        sweep_size: Number of sweeps for each step along the chain. Defaults to the number of sites in the Hilbert space.
                This is equivalent to subsampling the Markov chain.
        reset_chains: If True, resets the chain state when `reset` is called on every new sampling (default = False).
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the states sampled (default = np.float64).

    Examples:
       Sampling from a RBM machine in a 1D lattice of spin 1/2

       >>> import netket as nk
       >>>
       >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
       >>> hi=nk.hilbert.Spin(s=0.5, N=g.n_nodes)
       >>>
       >>> # Transverse-field Ising Hamiltonian
       >>> ha = nk.operator.Ising(hilbert=hi, h=1.0, graph=g)
       >>>
       >>> # Construct a MetropolisHamiltonian Sampler
       >>> sa = nk.sampler.MetropolisHamiltonian(hi, hamiltonian=ha)
       >>> print(sa)
       MetropolisSampler(rule = HamiltonianRuleNumba(operator=Ising(J=1.0, h=1.0; dim=100)), n_chains = 16, sweep_size = 100, reset_chains = False, machine_power = 2, dtype = <class 'float'>)
    """
    from .rules import HamiltonianRule

    rule = HamiltonianRule(hamiltonian)
    return MetropolisSampler(hilbert, rule, **kwargs)


def MetropolisGaussian(hilbert, sigma=1.0, **kwargs) -> MetropolisSampler:
    """This sampler acts on all particle positions simultaneously
    and proposes a new state according to a Gaussian distribution
    with width `sigma`.

    Args:
        hilbert: The continuous Hilbert space to sample.
        sigma: The width of the Gaussian proposal distribution (default = 1.0).
        n_chains: The total number of independent Markov chains across all MPI ranks. Either specify this or `n_chains_per_rank`.
        n_chains_per_rank: Number of independent chains on every MPI rank (default = 16).
        sweep_size: Number of sweeps for each step along the chain. Defaults to the number of sites in the Hilbert space.
                This is equivalent to subsampling the Markov chain.
        reset_chains: If True, resets the chain state when `reset` is called on every new sampling (default = False).
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the states sampled (default = np.float64).
    """
    if not isinstance(hilbert, ContinuousHilbert):
        raise ValueError("This sampler only works for Continuous Hilbert spaces.")

    from .rules import GaussianRule

    rule = GaussianRule(sigma)
    return MetropolisSampler(hilbert, rule, **kwargs)


def MetropolisAdjustedLangevin(
    hilbert, dt=0.001, chunk_size=None, **kwargs
) -> MetropolisSampler:
    r"""This sampler acts on all particle positions simultaneously
    and takes a Langevin step [1]:

    .. math::
       x_{t+dt} = x_t + dt \nabla_x \log p(x) \vert_{x=x_t} + \sqrt{2 dt}\eta,

    where  :math:`\eta` is normal distributed noise :math:`\eta \sim \mathcal{N}(0,1)`.
    This sampler only works for continuous Hilbert spaces.

    [1]: https://en.wikipedia.org/wiki/Metropolis-adjusted_Langevin_algorithm

    Args:
        hilbert: The continuous Hilbert space to sample.
        dt: Time step size for the Langevin dynamics (noise with variance 2*dt).
        chunk_size: Chunk size to compute the gradients of the log probability.
        n_chains: The total number of independent Markov chains across all MPI ranks. Either specify this or `n_chains_per_rank`.
        n_chains_per_rank: Number of independent chains on every MPI rank (default = 16).
        sweep_size: Number of sweeps for each step along the chain. Defaults to the number of sites in the Hilbert space.
                This is equivalent to subsampling the Markov chain.
        reset_chains: If True, resets the chain state when `reset` is called on every new sampling (default = False).
        machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
        dtype: The dtype of the states sampled (default = np.float64).
    """
    if not isinstance(hilbert, ContinuousHilbert):
        raise ValueError("This sampler only works for Continuous Hilbert spaces.")

    from .rules import LangevinRule

    rule = LangevinRule(dt=dt, chunk_size=chunk_size)
    return MetropolisSampler(hilbert, rule, **kwargs)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import jax.numpy as jnp

from netket.hilbert import DoubledHilbert, AbstractHilbert

from ._discrete_operator import DiscreteOperator


class AbstractSuperOperator(DiscreteOperator):
    """
    Generic base class for super-operators acting on the tensor product (DoubledHilbert)
    space ℋ⊗ℋ, where ℋ is the physical space.

    Behaves on :ref:`netket.vqs.VariationalMixedState` as normal operators behave
    on :ref:`netket.vqs.VariationalState`.
    Cannot be used to act upon pure states.
    """

    def __init__(self, hilbert):
        """
        Initialize a super-operator by passing it the physical hilbert space on which it acts.

        This init method constructs the doubled-hilbert space and pass it down to the fundamental
        abstractoperator.
        """
        super().__init__(DoubledHilbert(hilbert))

    @property
    def hilbert_physical(self) -> AbstractHilbert:
        """The physical hilbert space on which this super-operator acts."""
        return self.hilbert.physical

    def to_qobj(self):  # -> "qutip.Qobj"
        raise NotImplementedError("Superoperator to Qobj not yet implemented")

    def __matmul__(self, other):
        # Override DiscreteOperator to implement the Squared trick.
        # Should eventually remove it as well.
        if isinstance(other, np.ndarray) or isinstance(other, jnp.ndarray):
            return self.apply(other)
        elif isinstance(other, AbstractSuperOperator):
            if self == other and self.is_hermitian:
                from ._lazy import Squared

                return Squared(self)
            else:
                return self._op__matmul__(other)
        else:
            return NotImplemented

    def __rmatmul__(self, other):
        # override DiscreteOperator to implement the Squared trick.
        # Should eventually remove it as well.
        if isinstance(other, np.ndarray) or isinstance(other, jnp.ndarray):
            return NotImplemented
        elif isinstance(other, AbstractSuperOperator):
            if self == other and self.is_hermitian:
                from ._lazy import Squared

                return Squared(self)
            else:
                return self._op__rmatmul__(other)
        else:
            return NotImplemented
